@@ -34,7 +34,7 @@ from electrum.bitcoin import is_address
 from electrum.util import block_explorer_URL
 from electrum.plugin import run_hook
 
-from .util import MyTreeView, import_meta_gui, export_meta_gui, webopen
+from .util import MyTreeView, webopen
 
 
 class ContactList(MyTreeView):
@@ -63,12 +63,6 @@ class ContactList(MyTreeView):
         self.parent.set_contact(text, user_role)
         self.update()
 
-    def import_contacts(self):
-        import_meta_gui(self.parent, _('contacts'), self.parent.contacts.import_file, self.update)
-
-    def export_contacts(self):
-        export_meta_gui(self.parent, _('contacts'), self.parent.contacts.export_file)
-
     def create_menu(self, position):
         menu = QMenu()
         idx = self.indexAt(position)
@@ -85,7 +79,7 @@ class ContactList(MyTreeView):
             column_title = self.model().horizontalHeaderItem(column).text()
             column_data = '\n'.join(self.model().itemFromIndex(s_idx).text()
                                     for s_idx in self.selected_in_column(column))
-            menu.addAction(_("Copy {}").format(column_title), lambda: self.parent.app.clipboard().setText(column_data))
+            menu.addAction(_("Copy {}").format(column_title), lambda: self.place_text_on_clipboard(column_data, title=column_title))
             if column in self.editable_columns:
                 item = self.model().itemFromIndex(idx)
                 if item.isEditable():
@@ -102,6 +96,8 @@ class ContactList(MyTreeView):
         menu.exec_(self.viewport().mapToGlobal(position))
 
     def update(self):
+        if self.maybe_defer_update():
+            return
         current_key = self.current_item_user_role(col=self.Columns.NAME)
         self.model().clear()
         self.update_headers(self.__class__.headers)
