@@ -24,6 +24,7 @@
 # SOFTWARE.
 
 import os
+import sys
 import datetime
 from datetime import date
 from typing import TYPE_CHECKING, Tuple, Dict
@@ -38,18 +39,21 @@ from PyQt5.QtWidgets import (QMenu, QHeaderView, QLabel, QMessageBox,
                              QPushButton, QComboBox, QVBoxLayout, QCalendarWidget,
                              QGridLayout)
 
-from electrum.address_synchronizer import TX_HEIGHT_LOCAL
+from electrum.address_synchronizer import TX_HEIGHT_LOCAL, TX_HEIGHT_FUTURE
 from electrum.i18n import _
 from electrum.util import (block_explorer_URL, profiler, TxMinedInfo,
-                           OrderedDictWithIndex, timestamp_to_datetime)
+                           OrderedDictWithIndex, timestamp_to_datetime,
+                           Satoshis, Fiat, format_time)
 from electrum.logging import get_logger, Logger
 
+from .custom_model import CustomNode, CustomModel
 from .util import (read_QIcon, MONOSPACE_FONT, Buttons, CancelButton, OkButton,
                    filename_field, MyTreeView, AcceptFileDragDrop, WindowModalDialog,
                    CloseButton, webopen)
 
 if TYPE_CHECKING:
     from electrum.wallet import Abstract_Wallet
+    from .main_window import ElectrumWindow
 
 
 _logger = get_logger(__name__)
@@ -76,15 +80,14 @@ TX_ICONS = [
 ]
 
 class HistoryColumns(IntEnum):
-    STATUS_ICON = 0
-    STATUS_TEXT = 1
-    DESCRIPTION = 2
-    COIN_VALUE = 3
-    RUNNING_COIN_BALANCE = 4
-    FIAT_VALUE = 5
-    FIAT_ACQ_PRICE = 6
-    FIAT_CAP_GAINS = 7
-    TXID = 8
+    STATUS = 0
+    DESCRIPTION = 1
+    AMOUNT = 2
+    BALANCE = 3
+    FIAT_VALUE = 4
+    FIAT_ACQ_PRICE = 5
+    FIAT_CAP_GAINS = 6
+    TXID = 7
 
 class HistorySortModel(QSortFilterProxyModel):
     def lessThan(self, source_left: QModelIndex, source_right: QModelIndex):
@@ -101,81 +104,83 @@ class HistorySortModel(QSortFilterProxyModel):
         except:
             return False
 
-class HistoryModel(QAbstractItemModel, Logger):
+def get_item_key(tx_item):
+    return tx_item.get('txid') or tx_item['payment_hash']
 
-    def __init__(self, parent):
-        QAbstractItemModel.__init__(self, parent)
-        Logger.__init__(self)
-        self.parent = parent
-        self.view = None  # type: HistoryList
-        self.transactions = OrderedDictWithIndex()
-        self.tx_status_cache = {}  # type: Dict[str, Tuple[int, str]]
-        self.summary = None
 
-    def set_view(self, history_list: 'HistoryList'):
-        # FIXME HistoryModel and HistoryList mutually depend on each other.
-        # After constructing both, this method needs to be called.
-        self.view = history_list  # type: HistoryList
-        self.set_visibility_of_columns()
+class HistoryNode(CustomNode):
 
-    def columnCount(self, parent: QModelIndex):
-        return len(HistoryColumns)
-
-    def rowCount(self, parent: QModelIndex):
-        return len(self.transactions)
-
-    def index(self, row: int, column: int, parent: QModelIndex):
-        return self.createIndex(row, column)
-
-    def data(self, index: QModelIndex, role: Qt.ItemDataRole) -> QVariant:
+    def get_data_for_role(self, index: QModelIndex, role: Qt.ItemDataRole) -> QVariant:
         # note: this method is performance-critical.
         # it is called a lot, and so must run extremely fast.
         assert index.isValid()
         col = index.column()
-        tx_item = self.transactions.value_from_pos(index.row())
-        tx_hash = tx_item['txid']
-        conf = tx_item['confirmations']
-        txpos = tx_item['txpos_in_block'] or 0
-        height = tx_item['height']
-        try:
-            status, status_str = self.tx_status_cache[tx_hash]
-        except KeyError:
-            tx_mined_info = self.tx_mined_info_from_tx_item(tx_item)
-            status, status_str = self.parent.wallet.get_tx_status(tx_hash, tx_mined_info)
+        window = self.model.parent
+        tx_item = self.get_data()
+        is_lightning = tx_item.get('lightning', False)
+        timestamp = tx_item['timestamp']
+        if is_lightning:
+            status = 0
+            if timestamp is None:
+                status_str = 'unconfirmed'
+            else:
+                status_str = format_time(int(timestamp))
+        else:
+            tx_hash = tx_item['txid']
+            conf = tx_item['confirmations']
+            try:
+                status, status_str = self.model.tx_status_cache[tx_hash]
+            except KeyError:
+                tx_mined_info = self.model.tx_mined_info_from_tx_item(tx_item)
+                status, status_str = window.wallet.get_tx_status(tx_hash, tx_mined_info)
+
         if role == Qt.UserRole:
             # for sorting
             d = {
-                HistoryColumns.STATUS_ICON:
-                    # height breaks ties for unverified txns
-                    # txpos breaks ties for verified same block txns
-                    (conf, -status, -height, -txpos),
-                HistoryColumns.STATUS_TEXT: status_str,
-                HistoryColumns.DESCRIPTION: tx_item['label'],
-                HistoryColumns.COIN_VALUE:  tx_item['value'].value,
-                HistoryColumns.RUNNING_COIN_BALANCE: tx_item['balance'].value,
+                HistoryColumns.STATUS:
+                    # respect sort order of self.transactions (wallet.get_full_history)
+                    -index.row(),
+                HistoryColumns.DESCRIPTION:
+                    tx_item['label'] if 'label' in tx_item else None,
+                HistoryColumns.AMOUNT:
+                    (tx_item['bc_value'].value if 'bc_value' in tx_item else 0)\
+                    + (tx_item['ln_value'].value if 'ln_value' in tx_item else 0),
+                HistoryColumns.BALANCE:
+                    (tx_item['balance'].value if 'balance' in tx_item else 0),
                 HistoryColumns.FIAT_VALUE:
                     tx_item['fiat_value'].value if 'fiat_value' in tx_item else None,
                 HistoryColumns.FIAT_ACQ_PRICE:
                     tx_item['acquisition_price'].value if 'acquisition_price' in tx_item else None,
                 HistoryColumns.FIAT_CAP_GAINS:
                     tx_item['capital_gain'].value if 'capital_gain' in tx_item else None,
-                HistoryColumns.TXID: tx_hash,
+                HistoryColumns.TXID: tx_hash if not is_lightning else None,
             }
             return QVariant(d[col])
         if role not in (Qt.DisplayRole, Qt.EditRole):
-            if col == HistoryColumns.STATUS_ICON and role == Qt.DecorationRole:
-                return QVariant(read_QIcon(TX_ICONS[status]))
-            elif col == HistoryColumns.STATUS_ICON and role == Qt.ToolTipRole:
-                return QVariant(str(conf) + _(" confirmation" + ("s" if conf != 1 else "")))
+            if col == HistoryColumns.STATUS and role == Qt.DecorationRole:
+                icon = "lightning" if is_lightning else TX_ICONS[status]
+                return QVariant(read_QIcon(icon))
+            elif col == HistoryColumns.STATUS and role == Qt.ToolTipRole:
+                if is_lightning:
+                    msg = 'lightning transaction'
+                else:  # on-chain
+                    if tx_item['height'] == TX_HEIGHT_LOCAL:
+                        # note: should we also explain double-spends?
+                        msg = _("This transaction is only available on your local machine.\n"
+                                "The currently connected server does not know about it.\n"
+                                "You can either broadcast it now, or simply remove it.")
+                    else:
+                        msg = str(conf) + _(" confirmation" + ("s" if conf != 1 else ""))
+                return QVariant(msg)
             elif col > HistoryColumns.DESCRIPTION and role == Qt.TextAlignmentRole:
                 return QVariant(Qt.AlignRight | Qt.AlignVCenter)
-            elif col != HistoryColumns.STATUS_TEXT and role == Qt.FontRole:
+            elif col > HistoryColumns.DESCRIPTION and role == Qt.FontRole:
                 monospace_font = QFont(MONOSPACE_FONT)
                 return QVariant(monospace_font)
-            elif col == HistoryColumns.DESCRIPTION and role == Qt.DecorationRole \
-                    and self.parent.wallet.invoices.paid.get(tx_hash):
-                return QVariant(read_QIcon("seal"))
-            elif col in (HistoryColumns.DESCRIPTION, HistoryColumns.COIN_VALUE) \
+            #elif col == HistoryColumns.DESCRIPTION and role == Qt.DecorationRole and not is_lightning\
+            #        and self.parent.wallet.invoices.paid.get(tx_hash):
+            #    return QVariant(read_QIcon("seal"))
+            elif col in (HistoryColumns.DESCRIPTION, HistoryColumns.AMOUNT) \
                     and role == Qt.ForegroundRole and tx_item['value'].value < 0:
                 red_brush = QBrush(QColor("#BC1E1E"))
                 return QVariant(red_brush)
@@ -184,79 +189,144 @@ class HistoryModel(QAbstractItemModel, Logger):
                 blue_brush = QBrush(QColor("#1E1EFF"))
                 return QVariant(blue_brush)
             return QVariant()
-        if col == HistoryColumns.STATUS_TEXT:
+        if col == HistoryColumns.STATUS:
             return QVariant(status_str)
-        elif col == HistoryColumns.DESCRIPTION:
+        elif col == HistoryColumns.DESCRIPTION and 'label' in tx_item:
             return QVariant(tx_item['label'])
-        elif col == HistoryColumns.COIN_VALUE:
-            value = tx_item['value'].value
-            v_str = self.parent.format_amount(value, is_diff=True, whitespaces=True)
+        elif col == HistoryColumns.AMOUNT:
+            bc_value = tx_item['bc_value'].value if 'bc_value' in tx_item else 0
+            ln_value = tx_item['ln_value'].value if 'ln_value' in tx_item else 0
+            value = bc_value + ln_value
+            v_str = window.format_amount(value, is_diff=True, whitespaces=True)
             return QVariant(v_str)
-        elif col == HistoryColumns.RUNNING_COIN_BALANCE:
+        elif col == HistoryColumns.BALANCE:
             balance = tx_item['balance'].value
-            balance_str = self.parent.format_amount(balance, whitespaces=True)
+            balance_str = window.format_amount(balance, whitespaces=True)
             return QVariant(balance_str)
         elif col == HistoryColumns.FIAT_VALUE and 'fiat_value' in tx_item:
-            value_str = self.parent.fx.format_fiat(tx_item['fiat_value'].value)
+            value_str = window.fx.format_fiat(tx_item['fiat_value'].value)
             return QVariant(value_str)
         elif col == HistoryColumns.FIAT_ACQ_PRICE and \
                 tx_item['value'].value < 0 and 'acquisition_price' in tx_item:
             # fixme: should use is_mine
             acq = tx_item['acquisition_price'].value
-            return QVariant(self.parent.fx.format_fiat(acq))
+            return QVariant(window.fx.format_fiat(acq))
         elif col == HistoryColumns.FIAT_CAP_GAINS and 'capital_gain' in tx_item:
             cg = tx_item['capital_gain'].value
-            return QVariant(self.parent.fx.format_fiat(cg))
+            return QVariant(window.fx.format_fiat(cg))
         elif col == HistoryColumns.TXID:
-            return QVariant(tx_hash)
+            return QVariant(tx_hash) if not is_lightning else QVariant('')
         return QVariant()
 
-    def parent(self, index: QModelIndex):
-        return QModelIndex()
 
-    def hasChildren(self, index: QModelIndex):
-        return not index.isValid()
+class HistoryModel(CustomModel, Logger):
 
-    def update_label(self, row):
-        tx_item = self.transactions.value_from_pos(row)
-        tx_item['label'] = self.parent.wallet.get_label(tx_item['txid'])
-        topLeft = bottomRight = self.createIndex(row, 2)
+    def __init__(self, parent: 'ElectrumWindow'):
+        CustomModel.__init__(self, parent, len(HistoryColumns))
+        Logger.__init__(self)
+        self.parent = parent
+        self.view = None  # type: HistoryList
+        self.transactions = OrderedDictWithIndex()
+        self.tx_status_cache = {}  # type: Dict[str, Tuple[int, str]]
+
+    def set_view(self, history_list: 'HistoryList'):
+        # FIXME HistoryModel and HistoryList mutually depend on each other.
+        # After constructing both, this method needs to be called.
+        self.view = history_list  # type: HistoryList
+        self.set_visibility_of_columns()
+
+    def update_label(self, index):
+        tx_item = index.internalPointer().get_data()
+        tx_item['label'] = self.parent.wallet.get_label(get_item_key(tx_item))
+        topLeft = bottomRight = self.createIndex(index.row(), HistoryColumns.DESCRIPTION)
         self.dataChanged.emit(topLeft, bottomRight, [Qt.DisplayRole])
+        self.parent.utxo_list.update()
 
     def get_domain(self):
-        '''Overridden in address_dialog.py'''
+        """Overridden in address_dialog.py"""
         return self.parent.wallet.get_addresses()
 
+    def should_include_lightning_payments(self) -> bool:
+        """Overridden in address_dialog.py"""
+        return True
+
     @profiler
     def refresh(self, reason: str):
         self.logger.info(f"refreshing... reason: {reason}")
         assert self.parent.gui_thread == threading.current_thread(), 'must be called from GUI thread'
         assert self.view, 'view not set'
+        if self.view.maybe_defer_update():
+            return
         selected = self.view.selectionModel().currentIndex()
         selected_row = None
         if selected:
             selected_row = selected.row()
         fx = self.parent.fx
         if fx: fx.history_used_spot = False
-        r = self.parent.wallet.get_full_history(domain=self.get_domain(), from_timestamp=None, to_timestamp=None, fx=fx)
+        wallet = self.parent.wallet
         self.set_visibility_of_columns()
-        if r['transactions'] == list(self.transactions.values()):
+        transactions = wallet.get_full_history(
+            self.parent.fx,
+            onchain_domain=self.get_domain(),
+            include_lightning=self.should_include_lightning_payments())
+        if transactions == list(self.transactions.values()):
             return
-        old_length = len(self.transactions)
+        old_length = self._root.childCount()
         if old_length != 0:
             self.beginRemoveRows(QModelIndex(), 0, old_length)
             self.transactions.clear()
+            self._root = HistoryNode(self, None)
             self.endRemoveRows()
-        self.beginInsertRows(QModelIndex(), 0, len(r['transactions'])-1)
-        for tx_item in r['transactions']:
-            txid = tx_item['txid']
-            self.transactions[txid] = tx_item
+        parents = {}
+        for tx_item in transactions.values():
+            node = HistoryNode(self, tx_item)
+            group_id = tx_item.get('group_id')
+            if group_id is None:
+                self._root.addChild(node)
+            else:
+                parent = parents.get(group_id)
+                if parent is None:
+                    # create parent if it does not exist
+                    self._root.addChild(node)
+                    parents[group_id] = node
+                else:
+                    # if parent has no children, create two children
+                    if parent.childCount() == 0:
+                        child_data = dict(parent.get_data())
+                        node1 = HistoryNode(self, child_data)
+                        parent.addChild(node1)
+                        parent._data['label'] = child_data.get('group_label')
+                        parent._data['bc_value'] = child_data.get('bc_value', Satoshis(0))
+                        parent._data['ln_value'] = child_data.get('ln_value', Satoshis(0))
+                    # add child to parent
+                    parent.addChild(node)
+                    # update parent data
+                    parent._data['balance'] = tx_item['balance']
+                    parent._data['value'] += tx_item['value']
+                    if 'group_label' in tx_item:
+                        parent._data['label'] = tx_item['group_label']
+                    if 'bc_value' in tx_item:
+                        parent._data['bc_value'] += tx_item['bc_value']
+                    if 'ln_value' in tx_item:
+                        parent._data['ln_value'] += tx_item['ln_value']
+                    if 'fiat_value' in tx_item:
+                        parent._data['fiat_value'] += tx_item['fiat_value']
+                    if tx_item.get('txid') == group_id:
+                        parent._data['lightning'] = False
+                        parent._data['txid'] = tx_item['txid']
+                        parent._data['timestamp'] = tx_item['timestamp']
+                        parent._data['height'] = tx_item['height']
+                        parent._data['confirmations'] = tx_item['confirmations']
+
+        new_length = self._root.childCount()
+        self.beginInsertRows(QModelIndex(), 0, new_length-1)
+        self.transactions = transactions
         self.endInsertRows()
+
         if selected_row:
             self.view.selectionModel().select(self.createIndex(selected_row, 0), QItemSelectionModel.Rows | QItemSelectionModel.SelectCurrent)
         self.view.filter()
-        # update summary
-        self.summary = r['summary']
+        # update time filter
         if not self.view.years and self.transactions:
             start_date = date.today()
             end_date = date.today()
@@ -268,8 +338,9 @@ class HistoryModel(QAbstractItemModel, Logger):
         # update tx_status_cache
         self.tx_status_cache.clear()
         for txid, tx_item in self.transactions.items():
-            tx_mined_info = self.tx_mined_info_from_tx_item(tx_item)
-            self.tx_status_cache[txid] = self.parent.wallet.get_tx_status(txid, tx_mined_info)
+            if not tx_item.get('lightning', False):
+                tx_mined_info = self.tx_mined_info_from_tx_item(tx_item)
+                self.tx_status_cache[txid] = self.parent.wallet.get_tx_status(txid, tx_mined_info)
 
     def set_visibility_of_columns(self):
         def set_visible(col: int, b: bool):
@@ -283,8 +354,8 @@ class HistoryModel(QAbstractItemModel, Logger):
         set_visible(HistoryColumns.FIAT_ACQ_PRICE, history and cap_gains)
         set_visible(HistoryColumns.FIAT_CAP_GAINS, history and cap_gains)
 
-    def update_fiat(self, row, idx):
-        tx_item = self.transactions.value_from_pos(row)
+    def update_fiat(self, idx):
+        tx_item = idx.internalPointer().get_data()
         key = tx_item['txid']
         fee = tx_item.get('fee')
         value = tx_item['value'].value
@@ -311,6 +382,8 @@ class HistoryModel(QAbstractItemModel, Logger):
 
     def on_fee_histogram(self):
         for tx_hash, tx_item in list(self.transactions.items()):
+            if tx_item.get('lightning'):
+                continue
             tx_mined_info = self.tx_mined_info_from_tx_item(tx_item)
             if tx_mined_info.conf > 0:
                 # note: we could actually break here if we wanted to rely on the order of txns in self.transactions
@@ -330,11 +403,10 @@ class HistoryModel(QAbstractItemModel, Logger):
             fiat_acq_title = '%s '%fx.ccy + _('Acquisition price')
             fiat_cg_title =  '%s '%fx.ccy + _('Capital Gains')
         return {
-            HistoryColumns.STATUS_ICON: '',
-            HistoryColumns.STATUS_TEXT: _('Date'),
+            HistoryColumns.STATUS: _('Date'),
             HistoryColumns.DESCRIPTION: _('Description'),
-            HistoryColumns.COIN_VALUE: _('Amount'),
-            HistoryColumns.RUNNING_COIN_BALANCE: _('Balance'),
+            HistoryColumns.AMOUNT: _('Amount'),
+            HistoryColumns.BALANCE: _('Balance'),
             HistoryColumns.FIAT_VALUE: fiat_title,
             HistoryColumns.FIAT_ACQ_PRICE: fiat_acq_title,
             HistoryColumns.FIAT_CAP_GAINS: fiat_cg_title,
@@ -355,14 +427,14 @@ class HistoryModel(QAbstractItemModel, Logger):
         return tx_mined_info
 
 class HistoryList(MyTreeView, AcceptFileDragDrop):
-    filter_columns = [HistoryColumns.STATUS_TEXT,
+    filter_columns = [HistoryColumns.STATUS,
                       HistoryColumns.DESCRIPTION,
-                      HistoryColumns.COIN_VALUE,
+                      HistoryColumns.AMOUNT,
                       HistoryColumns.TXID]
 
     def tx_item_from_proxy_row(self, proxy_row):
         hm_idx = self.model().mapToSource(self.model().index(proxy_row, 0))
-        return self.hm.transactions.value_from_pos(hm_idx.row())
+        return hm_idx.internalPointer().get_data()
 
     def should_hide(self, proxy_row):
         if self.start_timestamp and self.end_timestamp:
@@ -376,12 +448,11 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
 
     def __init__(self, parent, model: HistoryModel):
         super().__init__(parent, self.create_menu, stretch_column=HistoryColumns.DESCRIPTION)
+        self.config = parent.config
         self.hm = model
         self.proxy = HistorySortModel(self)
         self.proxy.setSourceModel(model)
         self.setModel(self.proxy)
-
-        self.config = parent.config
         AcceptFileDragDrop.__init__(self, ".txn")
         self.setSortingEnabled(True)
         self.start_timestamp = None
@@ -389,14 +460,17 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
         self.years = []
         self.create_toolbar_buttons()
         self.wallet = self.parent.wallet  # type: Abstract_Wallet
-        self.sortByColumn(HistoryColumns.STATUS_ICON, Qt.AscendingOrder)
+        self.sortByColumn(HistoryColumns.STATUS, Qt.AscendingOrder)
         self.editable_columns |= {HistoryColumns.FIAT_VALUE}
-
+        self.setRootIsDecorated(True)
         self.header().setStretchLastSection(False)
         for col in HistoryColumns:
             sm = QHeaderView.Stretch if col == self.stretch_column else QHeaderView.ResizeToContents
             self.header().setSectionResizeMode(col, sm)
 
+    def update(self):
+        self.hm.refresh('HistoryList.update()')
+
     def format_date(self, d):
         return str(datetime.date(d.year, d.month, d.day)) if d else _('None')
 
@@ -472,7 +546,7 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
             return datetime.datetime(date.year, date.month, date.day)
 
     def show_summary(self):
-        h = self.model().sourceModel().summary
+        h = self.parent.wallet.get_detailed_history()['summary']
         if not h:
             self.parent.show_message(_("Nothing to summarize."))
             return
@@ -524,18 +598,18 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
 
     def on_edited(self, index, user_role, text):
         index = self.model().mapToSource(index)
-        row, column = index.row(), index.column()
-        tx_item = self.hm.transactions.value_from_pos(row)
-        key = tx_item['txid']
+        tx_item = index.internalPointer().get_data()
+        column = index.column()
+        key = get_item_key(tx_item)
         if column == HistoryColumns.DESCRIPTION:
             if self.wallet.set_label(key, text): #changed
-                self.hm.update_label(row)
+                self.hm.update_label(index)
                 self.parent.update_completions()
         elif column == HistoryColumns.FIAT_VALUE:
             self.wallet.set_fiat_value(key, self.parent.fx.ccy, text, self.parent.fx, tx_item['value'].value)
             value = tx_item['value'].value
             if value is not None:
-                self.hm.update_fiat(row, index)
+                self.hm.update_fiat(index)
         else:
             assert False
 
@@ -547,14 +621,34 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
         if self.hm.flags(self.model().mapToSource(idx)) & Qt.ItemIsEditable:
             super().mouseDoubleClickEvent(event)
         else:
-            self.show_transaction(tx_item['txid'])
+            if tx_item.get('lightning'):
+                if tx_item['type'] == 'payment':
+                    self.parent.show_lightning_transaction(tx_item)
+                return
+            tx_hash = tx_item['txid']
+            tx = self.wallet.db.get_transaction(tx_hash)
+            if not tx:
+                return
+            self.show_transaction(tx_item, tx)
 
-    def show_transaction(self, tx_hash):
-        tx = self.wallet.db.get_transaction(tx_hash)
-        if not tx:
-            return
+    def show_transaction(self, tx_item, tx):
+        tx_hash = tx_item['txid']
         label = self.wallet.get_label(tx_hash) or None # prefer 'None' if not defined (force tx dialog to hide Description field if missing)
-        self.parent.show_transaction(tx, label)
+        self.parent.show_transaction(tx, tx_desc=label)
+
+    def add_copy_menu(self, menu, idx):
+        cc = menu.addMenu(_("Copy"))
+        for column in HistoryColumns:
+            if self.isColumnHidden(column):
+                continue
+            column_title = self.hm.headerData(column, Qt.Horizontal, Qt.DisplayRole)
+            idx2 = idx.sibling(idx.row(), column)
+            column_data = (self.hm.data(idx2, Qt.DisplayRole).value() or '').strip()
+            cc.addAction(
+                column_title,
+                lambda text=column_data, title=column_title:
+                self.place_text_on_clipboard(text, title=title))
+        return cc
 
     def create_menu(self, position: QPoint):
         org_idx: QModelIndex = self.indexAt(position)
@@ -562,57 +656,60 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
         if not idx.isValid():
             # can happen e.g. before list is populated for the first time
             return
-        tx_item = self.hm.transactions.value_from_pos(idx.row())
-        column = idx.column()
-        if column == HistoryColumns.STATUS_ICON:
-            column_title = _('Transaction ID')
-            column_data = tx_item['txid']
-        else:
-            column_title = self.hm.headerData(column, Qt.Horizontal, Qt.DisplayRole)
-            column_data = self.hm.data(idx, Qt.DisplayRole).value()
+        tx_item = idx.internalPointer().get_data()
+        if tx_item.get('lightning') and tx_item['type'] == 'payment':
+            menu = QMenu()
+            menu.addAction(_("View Payment"), lambda: self.parent.show_lightning_transaction(tx_item))
+            cc = self.add_copy_menu(menu, idx)
+            cc.addAction(_("Payment Hash"), lambda: self.place_text_on_clipboard(tx_item['payment_hash'], title="Payment Hash"))
+            cc.addAction(_("Preimage"), lambda: self.place_text_on_clipboard(tx_item['preimage'], title="Preimage"))
+            menu.exec_(self.viewport().mapToGlobal(position))
+            return
         tx_hash = tx_item['txid']
-        tx = self.wallet.db.get_transaction(tx_hash)
+        if tx_item.get('lightning'):
+            tx = self.wallet.lnworker.lnwatcher.db.get_transaction(tx_hash)
+        else:
+            tx = self.wallet.db.get_transaction(tx_hash)
         if not tx:
             return
         tx_URL = block_explorer_URL(self.config, 'tx', tx_hash)
-        height = self.wallet.get_tx_height(tx_hash).height
-        is_relevant, is_mine, v, fee = self.wallet.get_wallet_delta(tx)
-        is_unconfirmed = height <= 0
-        pr_key = self.wallet.invoices.paid.get(tx_hash)
+        tx_details = self.wallet.get_tx_info(tx)
+        is_unconfirmed = tx_details.tx_mined_status.height <= 0
+        invoice_keys = self.wallet.get_relevant_invoice_keys_for_tx(tx)
         menu = QMenu()
-        if height == TX_HEIGHT_LOCAL:
+        if tx_details.can_remove:
             menu.addAction(_("Remove"), lambda: self.remove_local_tx(tx_hash))
-
-        amount_columns = [HistoryColumns.COIN_VALUE, HistoryColumns.RUNNING_COIN_BALANCE, HistoryColumns.FIAT_VALUE, HistoryColumns.FIAT_ACQ_PRICE, HistoryColumns.FIAT_CAP_GAINS]
-        if column in amount_columns:
-            column_data = column_data.strip()
-        menu.addAction(_("Copy {}").format(column_title), lambda: self.parent.app.clipboard().setText(column_data))
-
+        cc = self.add_copy_menu(menu, idx)
+        cc.addAction(_("Transaction ID"), lambda: self.place_text_on_clipboard(tx_hash, title="TXID"))
         for c in self.editable_columns:
             if self.isColumnHidden(c): continue
             label = self.hm.headerData(c, Qt.Horizontal, Qt.DisplayRole)
             # TODO use siblingAtColumn when min Qt version is >=5.11
             persistent = QPersistentModelIndex(org_idx.sibling(org_idx.row(), c))
             menu.addAction(_("Edit {}").format(label), lambda p=persistent: self.edit(QModelIndex(p)))
-        menu.addAction(_("Details"), lambda: self.show_transaction(tx_hash))
+        menu.addAction(_("View Transaction"), lambda: self.show_transaction(tx_item, tx))
+        channel_id = tx_item.get('channel_id')
+        if channel_id:
+            menu.addAction(_("View Channel"), lambda: self.parent.show_channel(bytes.fromhex(channel_id)))
         if is_unconfirmed and tx:
             # note: the current implementation of RBF *needs* the old tx fee
-            rbf = is_mine and not tx.is_final() and fee is not None
-            if rbf:
+            if tx_details.can_bump and tx_details.fee is not None:
                 menu.addAction(_("Increase fee"), lambda: self.parent.bump_fee_dialog(tx))
             else:
                 child_tx = self.wallet.cpfp(tx, 0)
                 if child_tx:
                     menu.addAction(_("Child pays for parent"), lambda: self.parent.cpfp(tx, child_tx))
-        if pr_key:
-            menu.addAction(read_QIcon("seal"), _("View invoice"), lambda: self.parent.show_invoice(pr_key))
+        for key in invoice_keys:
+            invoice = self.parent.wallet.get_invoice(key)
+            if invoice:
+                menu.addAction(_("View invoice"), lambda: self.parent.show_onchain_invoice(invoice))
         if tx_URL:
             menu.addAction(_("View on block explorer"), lambda: webopen(tx_URL))
         menu.exec_(self.viewport().mapToGlobal(position))
 
-    def remove_local_tx(self, delete_tx):
-        to_delete = {delete_tx}
-        to_delete |= self.wallet.get_depending_transactions(delete_tx)
+    def remove_local_tx(self, tx_hash: str):
+        to_delete = {tx_hash}
+        to_delete |= self.wallet.get_depending_transactions(tx_hash)
         question = _("Are you sure you want to remove this transaction?")
         if len(to_delete) > 1:
             question = (_("Are you sure you want to remove this transaction and {} child transactions?")
@@ -622,7 +719,7 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
             return
         for tx in to_delete:
             self.wallet.remove_transaction(tx)
-        self.wallet.storage.write()
+        self.wallet.save_db()
         # need to update at least: history_list, utxo_list, address_list
         self.parent.need_update.set()
 
@@ -630,9 +727,12 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
         try:
             with open(fn) as f:
                 tx = self.parent.tx_from_text(f.read())
-                self.parent.save_transaction_into_wallet(tx)
         except IOError as e:
             self.parent.show_error(e)
+            return
+        if not tx:
+            return
+        self.parent.save_transaction_into_wallet(tx)
 
     def export_history_dialog(self):
         d = WindowModalDialog(self, _('Export History'))
@@ -661,11 +761,7 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
         self.parent.show_message(_("Your wallet history has been successfully exported."))
 
     def do_export_history(self, file_name, is_csv):
-        hist = self.wallet.get_full_history(domain=self.hm.get_domain(),
-                                            from_timestamp=None,
-                                            to_timestamp=None,
-                                            fx=self.parent.fx,
-                                            show_fees=True)
+        hist = self.wallet.get_detailed_history(fx=self.parent.fx)
         txns = hist['transactions']
         lines = []
         if is_csv:
@@ -673,7 +769,7 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
                 lines.append([item['txid'],
                               item.get('label', ''),
                               item['confirmations'],
-                              item['value'],
+                              item['bc_value'],
                               item.get('fiat_value', ''),
                               item.get('fee', ''),
                               item.get('fiat_fee', ''),
@@ -696,7 +792,7 @@ class HistoryList(MyTreeView, AcceptFileDragDrop):
                 from electrum.util import json_encode
                 f.write(json_encode(txns))
 
-    def text_txid_from_coordinate(self, row, col):
+    def get_text_and_userrole_from_coordinate(self, row, col):
         idx = self.model().mapToSource(self.model().index(row, col))
-        tx_item = self.hm.transactions.value_from_pos(idx.row())
-        return self.hm.data(idx, Qt.DisplayRole).value(), tx_item['txid']
+        tx_item = idx.internalPointer().get_data()
+        return self.hm.data(idx, Qt.DisplayRole).value(), get_item_key(tx_item)
