@@ -26,73 +26,137 @@
 
 from unicodedata import normalize
 import hashlib
-from typing import Tuple
+import re
+from typing import Tuple, TYPE_CHECKING, Union, Sequence, Optional, Dict, List, NamedTuple
+from functools import lru_cache
+from abc import ABC, abstractmethod
 
 from . import bitcoin, ecc, constants, bip32
-from .bitcoin import (deserialize_privkey, serialize_privkey,
-                      public_key_to_p2pkh)
+from .bitcoin import deserialize_privkey, serialize_privkey
+from .transaction import Transaction, PartialTransaction, PartialTxInput, PartialTxOutput, TxInput
 from .bip32 import (convert_bip32_path_to_list_of_uint32, BIP32_PRIME,
-                    is_xpub, is_xprv, BIP32Node)
-from .ecc import string_to_number, number_to_string
+                    is_xpub, is_xprv, BIP32Node, normalize_bip32_derivation,
+                    convert_bip32_intpath_to_strpath, is_xkey_consistent_with_key_origin_info)
+from .ecc import string_to_number
 from .crypto import (pw_decode, pw_encode, sha256, sha256d, PW_HASH_VERSION_LATEST,
-                     SUPPORTED_PW_HASH_VERSIONS, UnsupportedPasswordHashVersion)
+                     SUPPORTED_PW_HASH_VERSIONS, UnsupportedPasswordHashVersion, hash_160)
 from .util import (InvalidPassword, WalletFileException,
-                   BitcoinException, bh2u, bfh, inv_dict)
-from .mnemonic import Mnemonic, load_wordlist, seed_type, is_seed
+                   BitcoinException, bh2u, bfh, inv_dict, is_hex_str)
+from .mnemonic import Mnemonic, Wordlist, seed_type, is_seed
 from .plugin import run_hook
 from .logging import Logger
 
+if TYPE_CHECKING:
+    from .gui.qt.util import TaskThread
+    from .plugins.hw_wallet import HW_PluginBase, HardwareClientBase, HardwareHandlerBase
+    from .wallet_db import WalletDB
 
-class KeyStore(Logger):
+
+class KeyStore(Logger, ABC):
+    type: str
 
     def __init__(self):
         Logger.__init__(self)
+        self.is_requesting_to_be_rewritten_to_wallet_file = False  # type: bool
 
-    def has_seed(self):
+    def has_seed(self) -> bool:
         return False
 
-    def is_watching_only(self):
+    def is_watching_only(self) -> bool:
         return False
 
-    def can_import(self):
+    def can_import(self) -> bool:
         return False
 
     def get_type_text(self) -> str:
         return f'{self.type}'
 
+    @abstractmethod
     def may_have_password(self):
         """Returns whether the keystore can be encrypted with a password."""
-        raise NotImplementedError()
+        pass
 
-    def get_tx_derivations(self, tx):
+    def _get_tx_derivations(self, tx: 'PartialTransaction') -> Dict[str, Union[Sequence[int], str]]:
         keypairs = {}
         for txin in tx.inputs():
-            num_sig = txin.get('num_sig')
-            if num_sig is None:
-                continue
-            x_signatures = txin['signatures']
-            signatures = [sig for sig in x_signatures if sig]
-            if len(signatures) == num_sig:
-                # input is complete
-                continue
-            for k, x_pubkey in enumerate(txin['x_pubkeys']):
-                if x_signatures[k] is not None:
-                    # this pubkey already signed
-                    continue
-                derivation = self.get_pubkey_derivation(x_pubkey)
-                if not derivation:
-                    continue
-                keypairs[x_pubkey] = derivation
+            keypairs.update(self._get_txin_derivations(txin))
         return keypairs
 
-    def can_sign(self, tx):
-        if self.is_watching_only():
-            return False
-        return bool(self.get_tx_derivations(tx))
+    def _get_txin_derivations(self, txin: 'PartialTxInput') -> Dict[str, Union[Sequence[int], str]]:
+        if txin.is_complete():
+            return {}
+        keypairs = {}
+        for pubkey in txin.pubkeys:
+            if pubkey in txin.part_sigs:
+                # this pubkey already signed
+                continue
+            derivation = self.get_pubkey_derivation(pubkey, txin)
+            if not derivation:
+                continue
+            keypairs[pubkey.hex()] = derivation
+        return keypairs
 
-    def ready_to_sign(self):
+    def can_sign(self, tx: 'Transaction', *, ignore_watching_only=False) -> bool:
+        """Returns whether this keystore could sign *something* in this tx."""
+        if not ignore_watching_only and self.is_watching_only():
+            return False
+        if not isinstance(tx, PartialTransaction):
+            return False
+        return bool(self._get_tx_derivations(tx))
+
+    def can_sign_txin(self, txin: 'TxInput', *, ignore_watching_only=False) -> bool:
+        """Returns whether this keystore could sign this txin."""
+        if not ignore_watching_only and self.is_watching_only():
+            return False
+        if not isinstance(txin, PartialTxInput):
+            return False
+        return bool(self._get_txin_derivations(txin))
+
+    def ready_to_sign(self) -> bool:
         return not self.is_watching_only()
 
+    @abstractmethod
+    def dump(self) -> dict:
+        pass
+
+    @abstractmethod
+    def is_deterministic(self) -> bool:
+        pass
+
+    @abstractmethod
+    def sign_message(self, sequence: 'AddressIndexGeneric', message, password) -> bytes:
+        pass
+
+    @abstractmethod
+    def decrypt_message(self, sequence: 'AddressIndexGeneric', message, password) -> bytes:
+        pass
+
+    @abstractmethod
+    def sign_transaction(self, tx: 'PartialTransaction', password) -> None:
+        pass
+
+    @abstractmethod
+    def get_pubkey_derivation(self, pubkey: bytes,
+                              txinout: Union['PartialTxInput', 'PartialTxOutput'],
+                              *, only_der_suffix=True) \
+            -> Union[Sequence[int], str, None]:
+        """Returns either a derivation int-list if the pubkey can be HD derived from this keystore,
+        the pubkey itself (hex) if the pubkey belongs to the keystore but not HD derived,
+        or None if the pubkey is unrelated.
+        """
+        pass
+
+    def find_my_pubkey_in_txinout(
+            self, txinout: Union['PartialTxInput', 'PartialTxOutput'],
+            *, only_der_suffix: bool = False
+    ) -> Tuple[Optional[bytes], Optional[List[int]]]:
+        # note: we assume that this cosigner only has one pubkey in this txin/txout
+        for pubkey in txinout.bip32_paths:
+            path = self.get_pubkey_derivation(pubkey, txinout, only_der_suffix=only_der_suffix)
+            if path and not isinstance(path, (str, bytes)):
+                return pubkey, list(path)
+        return None, None
+
 
 class Software_KeyStore(KeyStore):
 
@@ -122,21 +186,25 @@ class Software_KeyStore(KeyStore):
         # Raise if password is not correct.
         self.check_password(password)
         # Add private keys
-        keypairs = self.get_tx_derivations(tx)
+        keypairs = self._get_tx_derivations(tx)
         for k, v in keypairs.items():
             keypairs[k] = self.get_private_key(v, password)
         # Sign
         if keypairs:
             tx.sign(keypairs)
 
+    @abstractmethod
     def update_password(self, old_password, new_password):
-        raise NotImplementedError()  # implemented by subclasses
+        pass
 
+    @abstractmethod
     def check_password(self, password):
-        raise NotImplementedError()  # implemented by subclasses
+        pass
 
-    def get_private_key(self, *args, **kwargs) -> Tuple[bytes, bool]:
-        raise NotImplementedError()  # implemented by subclasses
+    @abstractmethod
+    def get_private_key(self, sequence: 'AddressIndexGeneric', password) -> Tuple[bytes, bool]:
+        """Returns (privkey, is_compressed)"""
+        pass
 
 
 class Imported_KeyStore(Software_KeyStore):
@@ -146,14 +214,11 @@ class Imported_KeyStore(Software_KeyStore):
 
     def __init__(self, d):
         Software_KeyStore.__init__(self, d)
-        self.keypairs = d.get('keypairs', {})
+        self.keypairs = d.get('keypairs', {})  # type: Dict[str, str]
 
     def is_deterministic(self):
         return False
 
-    def get_master_public_key(self):
-        return None
-
     def dump(self):
         return {
             'type': self.type,
@@ -184,7 +249,7 @@ class Imported_KeyStore(Software_KeyStore):
     def delete_imported_key(self, key):
         self.keypairs.pop(key)
 
-    def get_private_key(self, pubkey, password):
+    def get_private_key(self, pubkey: str, password):
         sec = pw_decode(self.keypairs[pubkey], password, version=self.pw_hash_version)
         txin_type, privkey, compressed = deserialize_privkey(sec)
         # this checks the password
@@ -192,14 +257,10 @@ class Imported_KeyStore(Software_KeyStore):
             raise InvalidPassword()
         return privkey, compressed
 
-    def get_pubkey_derivation(self, x_pubkey):
-        if x_pubkey[0:2] in ['02', '03', '04']:
-            if x_pubkey in self.keypairs.keys():
-                return x_pubkey
-        elif x_pubkey[0:2] == 'fd':
-            addr = bitcoin.script_to_address(x_pubkey[2:])
-            if addr in self.addresses:
-                return self.addresses[addr].get('pubkey')
+    def get_pubkey_derivation(self, pubkey, txin, *, only_der_suffix=True):
+        if pubkey.hex() in self.keypairs:
+            return pubkey.hex()
+        return None
 
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
@@ -212,7 +273,6 @@ class Imported_KeyStore(Software_KeyStore):
         self.pw_hash_version = PW_HASH_VERSION_LATEST
 
 
-
 class Deterministic_KeyStore(Software_KeyStore):
 
     def __init__(self, d):
@@ -240,12 +300,18 @@ class Deterministic_KeyStore(Software_KeyStore):
     def is_watching_only(self):
         return not self.has_seed()
 
+    @abstractmethod
+    def format_seed(self, seed: str) -> str:
+        pass
+
     def add_seed(self, seed):
         if self.seed:
             raise Exception("a seed exists")
         self.seed = self.format_seed(seed)
 
     def get_seed(self, password):
+        if not self.has_seed():
+            raise Exception("This wallet has no seed words")
         return pw_decode(self.seed, password, version=self.pw_hash_version)
 
     def get_passphrase(self, password):
@@ -255,20 +321,171 @@ class Deterministic_KeyStore(Software_KeyStore):
             return ''
 
 
-class Xpub:
+class MasterPublicKeyMixin(ABC):
 
-    def __init__(self):
+    @abstractmethod
+    def get_master_public_key(self) -> str:
+        pass
+
+    @abstractmethod
+    def get_derivation_prefix(self) -> Optional[str]:
+        """Returns to bip32 path from some root node to self.xpub
+        Note that the return value might be None; if it is unknown.
+        """
+        pass
+
+    @abstractmethod
+    def get_root_fingerprint(self) -> Optional[str]:
+        """Returns the bip32 fingerprint of the top level node.
+        This top level node is the node at the beginning of the derivation prefix,
+        i.e. applying the derivation prefix to it will result self.xpub
+        Note that the return value might be None; if it is unknown.
+        """
+        pass
+
+    @abstractmethod
+    def get_fp_and_derivation_to_be_used_in_partial_tx(self, der_suffix: Sequence[int], *,
+                                                       only_der_suffix: bool = True) -> Tuple[bytes, Sequence[int]]:
+        """Returns fingerprint and derivation path corresponding to a derivation suffix.
+        The fingerprint is either the root fp or the intermediate fp, depending on what is available
+        and 'only_der_suffix', and the derivation path is adjusted accordingly.
+        """
+        pass
+
+    @abstractmethod
+    def derive_pubkey(self, for_change: int, n: int) -> bytes:
+        pass
+
+    def get_pubkey_derivation(self, pubkey: bytes,
+                              txinout: Union['PartialTxInput', 'PartialTxOutput'],
+                              *, only_der_suffix=True) \
+            -> Union[Sequence[int], str, None]:
+        def test_der_suffix_against_pubkey(der_suffix: Sequence[int], pubkey: bytes) -> bool:
+            if len(der_suffix) != 2:
+                return False
+            if pubkey != self.derive_pubkey(*der_suffix):
+                return False
+            return True
+
+        if pubkey not in txinout.bip32_paths:
+            return None
+        fp_found, path_found = txinout.bip32_paths[pubkey]
+        der_suffix = None
+        full_path = None
+        # try fp against our root
+        my_root_fingerprint_hex = self.get_root_fingerprint()
+        my_der_prefix_str = self.get_derivation_prefix()
+        ks_der_prefix = convert_bip32_path_to_list_of_uint32(my_der_prefix_str) if my_der_prefix_str else None
+        if (my_root_fingerprint_hex is not None and ks_der_prefix is not None and
+                fp_found.hex() == my_root_fingerprint_hex):
+            if path_found[:len(ks_der_prefix)] == ks_der_prefix:
+                der_suffix = path_found[len(ks_der_prefix):]
+                if not test_der_suffix_against_pubkey(der_suffix, pubkey):
+                    der_suffix = None
+        # try fp against our intermediate fingerprint
+        if (der_suffix is None and isinstance(self, Xpub) and
+                fp_found == self.get_bip32_node_for_xpub().calc_fingerprint_of_this_node()):
+            der_suffix = path_found
+            if not test_der_suffix_against_pubkey(der_suffix, pubkey):
+                der_suffix = None
+        if der_suffix is None:
+            return None
+        if ks_der_prefix is not None:
+            full_path = ks_der_prefix + list(der_suffix)
+        return der_suffix if only_der_suffix else full_path
+
+
+class Xpub(MasterPublicKeyMixin):
+
+    def __init__(self, *, derivation_prefix: str = None, root_fingerprint: str = None):
         self.xpub = None
         self.xpub_receive = None
         self.xpub_change = None
+        self._xpub_bip32_node = None  # type: Optional[BIP32Node]
+
+        # "key origin" info (subclass should persist these):
+        self._derivation_prefix = derivation_prefix  # type: Optional[str]
+        self._root_fingerprint = root_fingerprint  # type: Optional[str]
 
     def get_master_public_key(self):
         return self.xpub
 
-    def derive_pubkey(self, for_change, n):
+    def get_bip32_node_for_xpub(self) -> Optional[BIP32Node]:
+        if self._xpub_bip32_node is None:
+            if self.xpub is None:
+                return None
+            self._xpub_bip32_node = BIP32Node.from_xkey(self.xpub)
+        return self._xpub_bip32_node
+
+    def get_derivation_prefix(self) -> Optional[str]:
+        return self._derivation_prefix
+
+    def get_root_fingerprint(self) -> Optional[str]:
+        return self._root_fingerprint
+
+    def get_fp_and_derivation_to_be_used_in_partial_tx(self, der_suffix: Sequence[int], *,
+                                                       only_der_suffix: bool = True) -> Tuple[bytes, Sequence[int]]:
+        fingerprint_hex = self.get_root_fingerprint()
+        der_prefix_str = self.get_derivation_prefix()
+        if not only_der_suffix and fingerprint_hex is not None and der_prefix_str is not None:
+            # use root fp, and true full path
+            fingerprint_bytes = bfh(fingerprint_hex)
+            der_prefix_ints = convert_bip32_path_to_list_of_uint32(der_prefix_str)
+        else:
+            # use intermediate fp, and claim der suffix is the full path
+            fingerprint_bytes = self.get_bip32_node_for_xpub().calc_fingerprint_of_this_node()
+            der_prefix_ints = convert_bip32_path_to_list_of_uint32('m')
+        der_full = der_prefix_ints + list(der_suffix)
+        return fingerprint_bytes, der_full
+
+    def get_xpub_to_be_used_in_partial_tx(self, *, only_der_suffix: bool) -> str:
+        assert self.xpub
+        fp_bytes, der_full = self.get_fp_and_derivation_to_be_used_in_partial_tx(der_suffix=[],
+                                                                                 only_der_suffix=only_der_suffix)
+        bip32node = self.get_bip32_node_for_xpub()
+        depth = len(der_full)
+        child_number_int = der_full[-1] if len(der_full) >= 1 else 0
+        child_number_bytes = child_number_int.to_bytes(length=4, byteorder="big")
+        fingerprint = bytes(4) if depth == 0 else bip32node.fingerprint
+        bip32node = bip32node._replace(depth=depth,
+                                       fingerprint=fingerprint,
+                                       child_number=child_number_bytes)
+        return bip32node.to_xpub()
+
+    def add_key_origin_from_root_node(self, *, derivation_prefix: str, root_node: BIP32Node):
+        assert self.xpub
+        # try to derive ourselves from what we were given
+        child_node1 = root_node.subkey_at_private_derivation(derivation_prefix)
+        child_pubkey_bytes1 = child_node1.eckey.get_public_key_bytes(compressed=True)
+        child_node2 = self.get_bip32_node_for_xpub()
+        child_pubkey_bytes2 = child_node2.eckey.get_public_key_bytes(compressed=True)
+        if child_pubkey_bytes1 != child_pubkey_bytes2:
+            raise Exception("(xpub, derivation_prefix, root_node) inconsistency")
+        self.add_key_origin(derivation_prefix=derivation_prefix,
+                            root_fingerprint=root_node.calc_fingerprint_of_this_node().hex().lower())
+
+    def add_key_origin(self, *, derivation_prefix: str = None, root_fingerprint: str = None) -> None:
+        assert self.xpub
+        if not (root_fingerprint is None or (is_hex_str(root_fingerprint) and len(root_fingerprint) == 8)):
+            raise Exception("root fp must be 8 hex characters")
+        derivation_prefix = normalize_bip32_derivation(derivation_prefix)
+        if not is_xkey_consistent_with_key_origin_info(self.xpub,
+                                                       derivation_prefix=derivation_prefix,
+                                                       root_fingerprint=root_fingerprint):
+            raise Exception("xpub inconsistent with provided key origin info")
+        if root_fingerprint is not None:
+            self._root_fingerprint = root_fingerprint
+        if derivation_prefix is not None:
+            self._derivation_prefix = derivation_prefix
+        self.is_requesting_to_be_rewritten_to_wallet_file = True
+
+    @lru_cache(maxsize=None)
+    def derive_pubkey(self, for_change: int, n: int) -> bytes:
+        for_change = int(for_change)
+        assert for_change in (0, 1)
         xpub = self.xpub_change if for_change else self.xpub_receive
         if xpub is None:
-            rootnode = BIP32Node.from_xkey(self.xpub)
+            rootnode = self.get_bip32_node_for_xpub()
             xpub = rootnode.subkey_at_public_derivation((for_change,)).to_xpub()
             if for_change:
                 self.xpub_change = xpub
@@ -277,49 +494,17 @@ class Xpub:
         return self.get_pubkey_from_xpub(xpub, (n,))
 
     @classmethod
-    def get_pubkey_from_xpub(self, xpub, sequence):
+    def get_pubkey_from_xpub(self, xpub: str, sequence) -> bytes:
         node = BIP32Node.from_xkey(xpub).subkey_at_public_derivation(sequence)
-        return node.eckey.get_public_key_hex(compressed=True)
-
-    def get_xpubkey(self, c, i):
-        s = ''.join(map(lambda x: bitcoin.int_to_hex(x,2), (c, i)))
-        return 'ff' + bh2u(bitcoin.DecodeBase58Check(self.xpub)) + s
-
-    @classmethod
-    def parse_xpubkey(self, pubkey):
-        # type + xpub + derivation
-        assert pubkey[0:2] == 'ff'
-        pk = bfh(pubkey)
-        # xpub:
-        pk = pk[1:]
-        xkey = bitcoin.EncodeBase58Check(pk[0:78])
-        # derivation:
-        dd = pk[78:]
-        s = []
-        # FIXME: due to an oversight, levels in the derivation are only
-        # allocated 2 bytes, instead of 4 (in bip32)
-        while dd:
-            n = int(bitcoin.rev_hex(bh2u(dd[0:2])), 16)
-            dd = dd[2:]
-            s.append(n)
-        assert len(s) == 2
-        return xkey, s
-
-    def get_pubkey_derivation(self, x_pubkey):
-        if x_pubkey[0:2] != 'ff':
-            return
-        xpub, derivation = self.parse_xpubkey(x_pubkey)
-        if self.xpub != xpub:
-            return
-        return derivation
+        return node.eckey.get_public_key_bytes(compressed=True)
 
 
-class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
+class BIP32_KeyStore(Xpub, Deterministic_KeyStore):
 
     type = 'bip32'
 
     def __init__(self, d):
-        Xpub.__init__(self)
+        Xpub.__init__(self, derivation_prefix=d.get('derivation'), root_fingerprint=d.get('root_fingerprint'))
         Deterministic_KeyStore.__init__(self, d)
         self.xpub = d.get('xpub')
         self.xprv = d.get('xprv')
@@ -331,6 +516,8 @@ class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
         d = Deterministic_KeyStore.dump(self)
         d['xpub'] = self.xpub
         d['xprv'] = self.xprv
+        d['derivation'] = self.get_derivation_prefix()
+        d['root_fingerprint'] = self.get_root_fingerprint()
         return d
 
     def get_master_private_key(self, password):
@@ -338,7 +525,7 @@ class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
 
     def check_password(self, password):
         xprv = pw_decode(self.xprv, password, version=self.pw_hash_version)
-        if BIP32Node.from_xkey(xprv).chaincode != BIP32Node.from_xkey(self.xpub).chaincode:
+        if BIP32Node.from_xkey(xprv).chaincode != self.get_bip32_node_for_xpub().chaincode:
             raise InvalidPassword()
 
     def update_password(self, old_password, new_password):
@@ -359,30 +546,43 @@ class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
     def is_watching_only(self):
         return self.xprv is None
 
+    def add_xpub(self, xpub):
+        assert is_xpub(xpub)
+        self.xpub = xpub
+        root_fingerprint, derivation_prefix = bip32.root_fp_and_der_prefix_from_xkey(xpub)
+        self.add_key_origin(derivation_prefix=derivation_prefix, root_fingerprint=root_fingerprint)
+
     def add_xprv(self, xprv):
+        assert is_xprv(xprv)
         self.xprv = xprv
-        self.xpub = bip32.xpub_from_xprv(xprv)
+        self.add_xpub(bip32.xpub_from_xprv(xprv))
 
     def add_xprv_from_seed(self, bip32_seed, xtype, derivation):
         rootnode = BIP32Node.from_rootseed(bip32_seed, xtype=xtype)
         node = rootnode.subkey_at_private_derivation(derivation)
         self.add_xprv(node.to_xprv())
+        self.add_key_origin_from_root_node(derivation_prefix=derivation, root_node=rootnode)
 
-    def get_private_key(self, sequence, password):
+    def get_private_key(self, sequence: Sequence[int], password):
         xprv = self.get_master_private_key(password)
         node = BIP32Node.from_xkey(xprv).subkey_at_private_derivation(sequence)
         pk = node.eckey.get_secret_bytes()
         return pk, True
 
+    def get_keypair(self, sequence, password):
+        k, _ = self.get_private_key(sequence, password)
+        cK = ecc.ECPrivkey(k).get_public_key_bytes()
+        return cK, k
 
 
-class Old_KeyStore(Deterministic_KeyStore):
+class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
 
     type = 'old'
 
     def __init__(self, d):
         Deterministic_KeyStore.__init__(self, d)
         self.mpk = d.get('mpk')
+        self._root_fingerprint = None
 
     def get_hex_seed(self, password):
         return pw_decode(self.seed, password, version=self.pw_hash_version).encode('utf8')
@@ -439,29 +639,32 @@ class Old_KeyStore(Deterministic_KeyStore):
         return string_to_number(sha256d(("%d:%d:"%(n, for_change)).encode('ascii') + bfh(mpk)))
 
     @classmethod
-    def get_pubkey_from_mpk(self, mpk, for_change, n):
-        z = self.get_sequence(mpk, for_change, n)
+    def get_pubkey_from_mpk(cls, mpk, for_change, n) -> bytes:
+        z = cls.get_sequence(mpk, for_change, n)
         master_public_key = ecc.ECPubkey(bfh('04'+mpk))
-        public_key = master_public_key + z*ecc.generator()
-        return public_key.get_public_key_hex(compressed=False)
+        public_key = master_public_key + z*ecc.GENERATOR
+        return public_key.get_public_key_bytes(compressed=False)
 
-    def derive_pubkey(self, for_change, n):
+    @lru_cache(maxsize=None)
+    def derive_pubkey(self, for_change, n) -> bytes:
+        for_change = int(for_change)
+        assert for_change in (0, 1)
         return self.get_pubkey_from_mpk(self.mpk, for_change, n)
 
-    def get_private_key_from_stretched_exponent(self, for_change, n, secexp):
+    def _get_private_key_from_stretched_exponent(self, for_change, n, secexp):
         secexp = (secexp + self.get_sequence(self.mpk, for_change, n)) % ecc.CURVE_ORDER
-        pk = number_to_string(secexp, ecc.CURVE_ORDER)
+        pk = int.to_bytes(secexp, length=32, byteorder='big', signed=False)
         return pk
 
-    def get_private_key(self, sequence, password):
+    def get_private_key(self, sequence: Sequence[int], password):
         seed = self.get_hex_seed(password)
         secexp = self.stretch_key(seed)
-        self.check_seed(seed, secexp=secexp)
+        self._check_seed(seed, secexp=secexp)
         for_change, n = sequence
-        pk = self.get_private_key_from_stretched_exponent(for_change, n, secexp)
+        pk = self._get_private_key_from_stretched_exponent(for_change, n, secexp)
         return pk, False
 
-    def check_seed(self, seed, *, secexp=None):
+    def _check_seed(self, seed, *, secexp=None):
         if secexp is None:
             secexp = self.stretch_key(seed)
         master_private_key = ecc.ECPrivkey.from_secret_scalar(secexp)
@@ -471,36 +674,29 @@ class Old_KeyStore(Deterministic_KeyStore):
 
     def check_password(self, password):
         seed = self.get_hex_seed(password)
-        self.check_seed(seed)
+        self._check_seed(seed)
 
     def get_master_public_key(self):
         return self.mpk
 
-    def get_xpubkey(self, for_change, n):
-        s = ''.join(map(lambda x: bitcoin.int_to_hex(x,2), (for_change, n)))
-        return 'fe' + self.mpk + s
+    def get_derivation_prefix(self) -> str:
+        return 'm'
 
-    @classmethod
-    def parse_xpubkey(self, x_pubkey):
-        assert x_pubkey[0:2] == 'fe'
-        pk = x_pubkey[2:]
-        mpk = pk[0:128]
-        dd = pk[128:]
-        s = []
-        while dd:
-            n = int(bitcoin.rev_hex(dd[0:4]), 16)
-            dd = dd[4:]
-            s.append(n)
-        assert len(s) == 2
-        return mpk, s
+    def get_root_fingerprint(self) -> str:
+        if self._root_fingerprint is None:
+            master_public_key = ecc.ECPubkey(bfh('04'+self.mpk))
+            xfp = hash_160(master_public_key.get_public_key_bytes(compressed=True))[0:4]
+            self._root_fingerprint = xfp.hex().lower()
+        return self._root_fingerprint
 
-    def get_pubkey_derivation(self, x_pubkey):
-        if x_pubkey[0:2] != 'fe':
-            return
-        mpk, derivation = self.parse_xpubkey(x_pubkey)
-        if self.mpk != mpk:
-            return
-        return derivation
+    def get_fp_and_derivation_to_be_used_in_partial_tx(self, der_suffix: Sequence[int], *,
+                                                       only_der_suffix: bool = True) -> Tuple[bytes, Sequence[int]]:
+        fingerprint_hex = self.get_root_fingerprint()
+        der_prefix_str = self.get_derivation_prefix()
+        fingerprint_bytes = bfh(fingerprint_hex)
+        der_prefix_ints = convert_bip32_path_to_list_of_uint32(der_prefix_str)
+        der_full = der_prefix_ints + list(der_suffix)
+        return fingerprint_bytes, der_full
 
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
@@ -512,25 +708,24 @@ class Old_KeyStore(Deterministic_KeyStore):
         self.pw_hash_version = PW_HASH_VERSION_LATEST
 
 
-
-class Hardware_KeyStore(KeyStore, Xpub):
-    # Derived classes must set:
-    #   - device
-    #   - DEVICE_IDS
-    #   - wallet_type
+class Hardware_KeyStore(Xpub, KeyStore):
+    hw_type: str
+    device: str
+    plugin: 'HW_PluginBase'
+    thread: Optional['TaskThread'] = None
 
     type = 'hardware'
 
     def __init__(self, d):
-        Xpub.__init__(self)
+        Xpub.__init__(self, derivation_prefix=d.get('derivation'), root_fingerprint=d.get('root_fingerprint'))
         KeyStore.__init__(self)
         # Errors and other user interaction is done through the wallet's
         # handler.  The handler is per-window and preserved across
         # device reconnects
         self.xpub = d.get('xpub')
         self.label = d.get('label')
-        self.derivation = d.get('derivation')
-        self.handler = None
+        self.soft_device_id = d.get('soft_device_id')  # type: Optional[str]
+        self.handler = None  # type: Optional[HardwareHandlerBase]
         run_hook('init_keystore', self)
 
     def set_label(self, label):
@@ -550,8 +745,10 @@ class Hardware_KeyStore(KeyStore, Xpub):
             'type': self.type,
             'hw_type': self.hw_type,
             'xpub': self.xpub,
-            'derivation':self.derivation,
+            'derivation': self.get_derivation_prefix(),
+            'root_fingerprint': self.get_root_fingerprint(),
             'label':self.label,
+            'soft_device_id': self.soft_device_id,
         }
 
     def unpaired(self):
@@ -564,24 +761,17 @@ class Hardware_KeyStore(KeyStore, Xpub):
         called in any thread context.'''
         self.logger.info("paired")
 
-    def can_export(self):
-        return False
-
     def is_watching_only(self):
         '''The wallet is not watching-only; the user will be prompted for
         pin and passphrase as appropriate when needed.'''
         assert not self.has_seed()
         return False
 
-    def get_password_for_storage_encryption(self):
-        from .storage import get_derivation_used_for_hw_device_encryption
+    def get_password_for_storage_encryption(self) -> str:
         client = self.plugin.get_client(self)
-        derivation = get_derivation_used_for_hw_device_encryption()
-        xpub = client.get_xpub(derivation, "standard")
-        password = self.get_pubkey_from_xpub(xpub, ())
-        return password
+        return client.get_password_for_storage_encryption()
 
-    def has_usable_connection_with_device(self):
+    def has_usable_connection_with_device(self) -> bool:
         if not hasattr(self, 'plugin'):
             return False
         client = self.plugin.get_client(self, force_pair=False)
@@ -592,6 +782,22 @@ class Hardware_KeyStore(KeyStore, Xpub):
     def ready_to_sign(self):
         return super().ready_to_sign() and self.has_usable_connection_with_device()
 
+    def opportunistically_fill_in_missing_info_from_device(self, client: 'HardwareClientBase'):
+        assert client is not None
+        if self._root_fingerprint is None:
+            self._root_fingerprint = client.request_root_fingerprint_from_device()
+            self.is_requesting_to_be_rewritten_to_wallet_file = True
+        if self.label != client.label():
+            self.label = client.label()
+            self.is_requesting_to_be_rewritten_to_wallet_file = True
+        if self.soft_device_id != client.get_soft_device_id():
+            self.soft_device_id = client.get_soft_device_id()
+            self.is_requesting_to_be_rewritten_to_wallet_file = True
+
+
+KeyStoreWithMPK = Union[KeyStore, MasterPublicKeyMixin]  # intersection really...
+AddressIndexGeneric = Union[Sequence[int], str]  # can be hex pubkey str
+
 
 def bip39_normalize_passphrase(passphrase):
     return normalize('NFKD', passphrase or '')
@@ -611,7 +817,7 @@ def bip39_is_checksum_valid(mnemonic: str) -> Tuple[bool, bool]:
     """
     words = [ normalize('NFKD', word) for word in mnemonic.split() ]
     words_len = len(words)
-    wordlist = load_wordlist("english.txt")
+    wordlist = Wordlist.from_file("english.txt")
     n = len(wordlist)
     i = 0
     words.reverse()
@@ -652,16 +858,17 @@ PURPOSE48_SCRIPT_TYPES_INV = inv_dict(PURPOSE48_SCRIPT_TYPES)
 
 def xtype_from_derivation(derivation: str) -> str:
     """Returns the script type to be used for this derivation."""
-    if derivation.startswith("m/84'"):
-        return 'p2wpkh'
-    elif derivation.startswith("m/49'"):
-        return 'p2wpkh-p2sh'
-    elif derivation.startswith("m/44'"):
-        return 'standard'
-    elif derivation.startswith("m/45'"):
-        return 'standard'
-
     bip32_indices = convert_bip32_path_to_list_of_uint32(derivation)
+    if len(bip32_indices) >= 1:
+        if bip32_indices[0] == 84 + BIP32_PRIME:
+            return 'p2wpkh'
+        elif bip32_indices[0] == 49 + BIP32_PRIME:
+            return 'p2wpkh-p2sh'
+        elif bip32_indices[0] == 44 + BIP32_PRIME:
+            return 'standard'
+        elif bip32_indices[0] == 45 + BIP32_PRIME:
+            return 'standard'
+
     if len(bip32_indices) >= 4:
         if bip32_indices[0] == 48 + BIP32_PRIME:
             # m / purpose' / coin_type' / account' / script_type' / change / address_index
@@ -672,46 +879,12 @@ def xtype_from_derivation(derivation: str) -> str:
     return 'standard'
 
 
-# extended pubkeys
-
-def is_xpubkey(x_pubkey):
-    return x_pubkey[0:2] == 'ff'
-
-
-def parse_xpubkey(x_pubkey):
-    assert x_pubkey[0:2] == 'ff'
-    return BIP32_KeyStore.parse_xpubkey(x_pubkey)
-
-
-def xpubkey_to_address(x_pubkey):
-    if x_pubkey[0:2] == 'fd':
-        address = bitcoin.script_to_address(x_pubkey[2:])
-        return x_pubkey, address
-    if x_pubkey[0:2] in ['02', '03', '04']:
-        pubkey = x_pubkey
-    elif x_pubkey[0:2] == 'ff':
-        xpub, s = BIP32_KeyStore.parse_xpubkey(x_pubkey)
-        pubkey = BIP32_KeyStore.get_pubkey_from_xpub(xpub, s)
-    elif x_pubkey[0:2] == 'fe':
-        mpk, s = Old_KeyStore.parse_xpubkey(x_pubkey)
-        pubkey = Old_KeyStore.get_pubkey_from_mpk(mpk, s[0], s[1])
-    else:
-        raise BitcoinException("Cannot parse pubkey. prefix: {}"
-                               .format(x_pubkey[0:2]))
-    if pubkey:
-        address = public_key_to_p2pkh(bfh(pubkey))
-    return pubkey, address
-
-def xpubkey_to_pubkey(x_pubkey):
-    pubkey, address = xpubkey_to_address(x_pubkey)
-    return pubkey
-
 hw_keystores = {}
 
 def register_keystore(hw_type, constructor):
     hw_keystores[hw_type] = constructor
 
-def hardware_keystore(d):
+def hardware_keystore(d) -> Hardware_KeyStore:
     hw_type = d['hw_type']
     if hw_type in hw_keystores:
         constructor = hw_keystores[hw_type]
@@ -719,8 +892,9 @@ def hardware_keystore(d):
     raise WalletFileException(f'unknown hardware type: {hw_type}. '
                               f'hw_keystores: {list(hw_keystores)}')
 
-def load_keystore(storage, name):
-    d = storage.get(name, {})
+def load_keystore(db: 'WalletDB', name: str) -> KeyStore:
+    d = db.get(name, {})
+    d = dict(d)  # convert to dict from StoredDict (see #6066)
     t = d.get('type')
     if not t:
         raise WalletFileException(
@@ -738,7 +912,7 @@ def load_keystore(storage, name):
 
 def is_old_mpk(mpk: str) -> bool:
     try:
-        int(mpk, 16)
+        int(mpk, 16)  # test if hex string
     except:
         return False
     if len(mpk) != 128:
@@ -772,16 +946,18 @@ def is_private_key_list(text, *, allow_spaces_inside_key=True, raise_on_error=Fa
                                  raise_on_error=raise_on_error))
 
 
-is_mpk = lambda x: is_old_mpk(x) or is_xpub(x)
-is_private = lambda x: is_seed(x) or is_xprv(x) or is_private_key_list(x)
-is_master_key = lambda x: is_old_mpk(x) or is_xprv(x) or is_xpub(x)
-is_private_key = lambda x: is_xprv(x) or is_private_key_list(x)
-is_bip32_key = lambda x: is_xprv(x) or is_xpub(x)
+def is_master_key(x):
+    return is_old_mpk(x) or is_bip32_key(x)
+
+
+def is_bip32_key(x):
+    return is_xprv(x) or is_xpub(x)
 
 
 def bip44_derivation(account_id, bip43_purpose=44):
     coin = constants.net.BIP44_COIN_TYPE
-    return "m/%d'/%d'/%d'" % (bip43_purpose, coin, int(account_id))
+    der = "m/%d'/%d'/%d'" % (bip43_purpose, coin, int(account_id))
+    return normalize_bip32_derivation(der)
 
 
 def purpose48_derivation(account_id: int, xtype: str) -> str:
@@ -792,7 +968,8 @@ def purpose48_derivation(account_id: int, xtype: str) -> str:
     script_type_int = PURPOSE48_SCRIPT_TYPES.get(xtype)
     if script_type_int is None:
         raise Exception('unknown xtype: {}'.format(xtype))
-    return "m/%d'/%d'/%d'/%d'" % (bip43_purpose, coin, account_id, script_type_int)
+    der = "m/%d'/%d'/%d'/%d'" % (bip43_purpose, coin, account_id, script_type_int)
+    return normalize_bip32_derivation(der)
 
 
 def from_seed(seed, passphrase, is_p2sh=False):
@@ -813,7 +990,7 @@ def from_seed(seed, passphrase, is_p2sh=False):
             xtype = 'p2wsh' if is_p2sh else 'p2wpkh'
         keystore.add_xprv_from_seed(bip32_seed, xtype, der)
     else:
-        raise BitcoinException('Unexpected seed type {}'.format(t))
+        raise BitcoinException('Unexpected seed type {}'.format(repr(t)))
     return keystore
 
 def from_private_key_list(text):
@@ -829,14 +1006,12 @@ def from_old_mpk(mpk):
 
 def from_xpub(xpub):
     k = BIP32_KeyStore({})
-    k.xpub = xpub
+    k.add_xpub(xpub)
     return k
 
 def from_xprv(xprv):
-    xpub = bip32.xpub_from_xprv(xprv)
     k = BIP32_KeyStore({})
-    k.xprv = xprv
-    k.xpub = xpub
+    k.add_xprv(xprv)
     return k
 
 def from_master_key(text):
