@@ -1,5 +1,6 @@
 import time
 from struct import pack
+from typing import Optional
 
 from electrum import ecc
 from electrum.i18n import _
@@ -7,10 +8,12 @@ from electrum.util import UserCancelled
 from electrum.keystore import bip39_normalize_passphrase
 from electrum.bip32 import BIP32Node, convert_bip32_path_to_list_of_uint32
 from electrum.logging import Logger
+from electrum.plugins.hw_wallet.plugin import HardwareClientBase, HardwareHandlerBase
 
 
 class GuiMixin(object):
     # Requires: self.proto, self.device
+    handler: Optional[HardwareHandlerBase]
 
     # ref: https://github.com/trezor/trezor-common/blob/44dfb07cfaafffada4b2ce0d15ba1d90d17cf35e/protob/types.proto#L89
     messages = {
@@ -46,6 +49,7 @@ class GuiMixin(object):
         return self.proto.ButtonAck()
 
     def callback_PinMatrixRequest(self, msg):
+        show_strength = True
         if msg.type == 2:
             msg = _("Enter a new PIN for your {}:")
         elif msg.type == 3:
@@ -53,7 +57,8 @@ class GuiMixin(object):
                      "NOTE: the positions of the numbers have changed!"))
         else:
             msg = _("Enter your current {} PIN:")
-        pin = self.handler.get_pin(msg.format(self.device))
+            show_strength = False
+        pin = self.handler.get_pin(msg.format(self.device), show_strength=show_strength)
         if len(pin) > 9:
             self.handler.show_error(_('The PIN cannot be longer than 9 characters.'))
             pin = ''  # to cancel below
@@ -96,10 +101,11 @@ class GuiMixin(object):
         return self.proto.WordAck(word=word)
 
 
-class SafeTClientBase(GuiMixin, Logger):
+class SafeTClientBase(HardwareClientBase, GuiMixin, Logger):
 
     def __init__(self, handler, plugin, proto):
         assert hasattr(self, 'tx_api')  # ProtocolMixin already constructed?
+        HardwareClientBase.__init__(self, plugin=plugin)
         self.proto = proto
         self.device = plugin.device
         self.handler = handler
@@ -114,11 +120,12 @@ class SafeTClientBase(GuiMixin, Logger):
         return "%s/%s" % (self.label(), self.features.device_id)
 
     def label(self):
-        '''The name given by the user to the device.'''
         return self.features.label
 
+    def get_soft_device_id(self):
+        return self.features.device_id
+
     def is_initialized(self):
-        '''True if initialized, False if wiped.'''
         return self.features.initialized
 
     def is_pairable(self):
