@@ -1,21 +1,21 @@
 import traceback
 import sys
-from typing import NamedTuple, Any
+from typing import NamedTuple, Any, Optional, Dict, Union, List, Tuple, TYPE_CHECKING
 
 from electrum.util import bfh, bh2u, versiontuple, UserCancelled, UserFacingException
-from electrum.bitcoin import TYPE_ADDRESS, TYPE_SCRIPT
 from electrum.bip32 import BIP32Node, convert_bip32_path_to_list_of_uint32 as parse_path
 from electrum import constants
 from electrum.i18n import _
 from electrum.plugin import Device
-from electrum.transaction import deserialize, Transaction
-from electrum.keystore import Hardware_KeyStore, is_xpubkey, parse_xpubkey
+from electrum.transaction import Transaction, PartialTransaction, PartialTxInput, PartialTxOutput
+from electrum.keystore import Hardware_KeyStore
 from electrum.base_wizard import ScriptTypeNotSupported, HWD_SETUP_NEW_WALLET
 from electrum.logging import get_logger
 
 from ..hw_wallet import HW_PluginBase
 from ..hw_wallet.plugin import (is_any_tx_output_on_change_branch, trezor_validate_op_return_output_and_get_data,
-                                LibraryFoundButUnusable, OutdatedHwFirmwareException)
+                                LibraryFoundButUnusable, OutdatedHwFirmwareException,
+                                get_xpubs_and_der_suffixes_from_txinout)
 
 _logger = get_logger(__name__)
 
@@ -28,19 +28,33 @@ try:
     from .clientbase import TrezorClientBase
 
     from trezorlib.messages import (
-        RecoveryDeviceType, HDNodeType, HDNodePathType,
+        Capability, BackupType, RecoveryDeviceType, HDNodeType, HDNodePathType,
         InputScriptType, OutputScriptType, MultisigRedeemScriptType,
         TxInputType, TxOutputType, TxOutputBinType, TransactionType, SignTx)
 
-    RECOVERY_TYPE_SCRAMBLED_WORDS = RecoveryDeviceType.ScrambledWords
-    RECOVERY_TYPE_MATRIX = RecoveryDeviceType.Matrix
+    from trezorlib.client import PASSPHRASE_ON_DEVICE
 
     TREZORLIB = True
 except Exception as e:
     _logger.exception('error importing trezorlib')
     TREZORLIB = False
 
-    RECOVERY_TYPE_SCRAMBLED_WORDS, RECOVERY_TYPE_MATRIX = range(2)
+    class _EnumMissing:
+        def __init__(self):
+            self.counter = 0
+            self.values = {}
+
+        def __getattr__(self, key):
+            if key not in self.values:
+                self.values[key] = self.counter
+                self.counter += 1
+            return self.values[key]
+
+    Capability = _EnumMissing()
+    BackupType = _EnumMissing()
+    RecoveryDeviceType = _EnumMissing()
+
+    PASSPHRASE_ON_DEVICE = object()
 
 
 # Trezor initialization methods
@@ -53,8 +67,7 @@ class TrezorKeyStore(Hardware_KeyStore):
     hw_type = 'trezor'
     device = TREZOR_PRODUCT_KEY
 
-    def get_derivation(self):
-        return self.derivation
+    plugin: 'TrezorPlugin'
 
     def get_client(self, force_pair=True):
         return self.plugin.get_client(self, force_pair)
@@ -64,7 +77,7 @@ class TrezorKeyStore(Hardware_KeyStore):
 
     def sign_message(self, sequence, message, password):
         client = self.get_client()
-        address_path = self.get_derivation() + "/%d/%d"%sequence
+        address_path = self.get_derivation_prefix() + "/%d/%d"%sequence
         msg_sig = client.sign_message(address_path, message)
         return msg_sig.signature
 
@@ -73,22 +86,13 @@ class TrezorKeyStore(Hardware_KeyStore):
             return
         # previous transactions used as inputs
         prev_tx = {}
-        # path of the xpubs that are involved
-        xpub_path = {}
         for txin in tx.inputs():
-            pubkeys, x_pubkeys = tx.get_sorted_pubkeys(txin)
-            tx_hash = txin['prevout_hash']
-            if txin.get('prev_tx') is None and not Transaction.is_segwit_input(txin):
-                raise UserFacingException(_('Offline signing with {} is not supported for legacy inputs.').format(self.device))
-            prev_tx[tx_hash] = txin['prev_tx']
-            for x_pubkey in x_pubkeys:
-                if not is_xpubkey(x_pubkey):
-                    continue
-                xpub, s = parse_xpubkey(x_pubkey)
-                if xpub == self.get_master_public_key():
-                    xpub_path[xpub] = self.get_derivation()
+            tx_hash = txin.prevout.txid.hex()
+            if txin.utxo is None:
+                raise UserFacingException(_('Missing previous tx.'))
+            prev_tx[tx_hash] = txin.utxo
 
-        self.plugin.sign_transaction(self, tx, prev_tx, xpub_path)
+        self.plugin.sign_transaction(self, tx, prev_tx)
 
 
 class TrezorInitSettings(NamedTuple):
@@ -97,6 +101,7 @@ class TrezorInitSettings(NamedTuple):
     pin_enabled: bool
     passphrase_enabled: bool
     recovery_type: Any = None
+    backup_type: int = BackupType.Bip39
     no_backup: bool = False
 
 
@@ -108,11 +113,11 @@ class TrezorPlugin(HW_PluginBase):
     #     wallet_class, types
 
     firmware_URL = 'https://wallet.trezor.io'
-    libraries_URL = 'https://github.com/trezor/python-trezor'
+    libraries_URL = 'https://pypi.org/project/trezor/'
     minimum_firmware = (1, 5, 2)
     keystore_class = TrezorKeyStore
-    minimum_library = (0, 11, 0)
-    maximum_library = (0, 12)
+    minimum_library = (0, 12, 0)
+    maximum_library = (0, 13)
     SUPPORTED_XTYPES = ('standard', 'p2wpkh-p2sh', 'p2wpkh', 'p2wsh-p2sh', 'p2wsh')
     DEVICE_IDS = (TREZOR_PRODUCT_KEY,)
 
@@ -172,11 +177,11 @@ class TrezorPlugin(HW_PluginBase):
         # note that this call can still raise!
         return TrezorClientBase(transport, handler, self)
 
-    def get_client(self, keystore, force_pair=True):
-        devmgr = self.device_manager()
-        handler = keystore.handler
-        with devmgr.hid_lock:
-            client = devmgr.client_for_keystore(self, handler, keystore, force_pair)
+    def get_client(self, keystore, force_pair=True, *,
+                   devices=None, allow_user_interaction=True) -> Optional['TrezorClientBase']:
+        client = super().get_client(keystore, force_pair,
+                                    devices=devices,
+                                    allow_user_interaction=allow_user_interaction)
         # returns the client for a given keystore. can use xpub
         if client:
             client.used()
@@ -187,14 +192,7 @@ class TrezorPlugin(HW_PluginBase):
 
     def initialize_device(self, device_id, wizard, handler):
         # Initialization method
-        msg = _("Choose how you want to initialize your {}.\n\n"
-                "The first two methods are secure as no secret information "
-                "is entered into your computer.\n\n"
-                "For the last two methods you input secrets on your keyboard "
-                "and upload them to your {}, and so you should "
-                "only do those on a computer you know to be trustworthy "
-                "and free of malware."
-        ).format(self.device, self.device)
+        msg = _("Choose how you want to initialize your {}.").format(self.device, self.device)
         choices = [
             # Must be short as QT doesn't word-wrap radio button text
             (TIM_NEW, _("Let the device generate a completely new seed randomly")),
@@ -222,13 +220,13 @@ class TrezorPlugin(HW_PluginBase):
             exit_code = 1
         except BaseException as e:
             self.logger.exception('')
-            handler.show_error(str(e))
+            handler.show_error(repr(e))
             exit_code = 1
         finally:
             wizard.loop.exit(exit_code)
 
     def _initialize_device(self, settings: TrezorInitSettings, method, device_id, wizard, handler):
-        if method == TIM_RECOVER and settings.recovery_type == RECOVERY_TYPE_SCRAMBLED_WORDS:
+        if method == TIM_RECOVER and settings.recovery_type == RecoveryDeviceType.ScrambledWords:
             handler.show_error(_(
                 "You will be asked to enter 24 words regardless of your "
                 "seed's actual length.  If you enter a word incorrectly or "
@@ -243,12 +241,13 @@ class TrezorPlugin(HW_PluginBase):
             raise Exception(_("The device was disconnected."))
 
         if method == TIM_NEW:
-            strength_from_word_count = {12: 128, 18: 192, 24: 256}
+            strength_from_word_count = {12: 128, 18: 192, 20: 128, 24: 256, 33: 256}
             client.reset_device(
                 strength=strength_from_word_count[settings.word_count],
                 passphrase_protection=settings.passphrase_enabled,
                 pin_protection=settings.pin_enabled,
                 label=settings.label,
+                backup_type=settings.backup_type,
                 no_backup=settings.no_backup)
         elif method == TIM_RECOVER:
             client.recover_device(
@@ -257,7 +256,7 @@ class TrezorPlugin(HW_PluginBase):
                 passphrase_protection=settings.passphrase_enabled,
                 pin_protection=settings.pin_enabled,
                 label=settings.label)
-            if settings.recovery_type == RECOVERY_TYPE_MATRIX:
+            if settings.recovery_type == RecoveryDeviceType.Matrix:
                 handler.close_matrix_dialog()
         else:
             raise RuntimeError("Unsupported recovery method")
@@ -274,12 +273,8 @@ class TrezorPlugin(HW_PluginBase):
         return HDNodePathType(node=node, address_n=address_n)
 
     def setup_device(self, device_info, wizard, purpose):
-        devmgr = self.device_manager()
         device_id = device_info.device.id_
-        client = devmgr.client_by_id(device_id)
-        if client is None:
-            raise UserFacingException(_('Failed to create a client for this device.') + '\n' +
-                                      _('Make sure it is in the correct state.'))
+        client = self.scan_and_create_client_for_device(device_id=device_id, wizard=wizard)
 
         if not client.is_uptodate():
             msg = (_('Outdated {} firmware for device labelled {}. Please '
@@ -287,20 +282,18 @@ class TrezorPlugin(HW_PluginBase):
                    .format(self.device, client.label(), self.firmware_URL))
             raise OutdatedHwFirmwareException(msg)
 
-        # fixme: we should use: client.handler = wizard
-        client.handler = self.create_handler(wizard)
         if not device_info.initialized:
             self.initialize_device(device_id, wizard, client.handler)
         is_creating_wallet = purpose == HWD_SETUP_NEW_WALLET
-        client.get_xpub('m', 'standard', creating=is_creating_wallet)
+        wizard.run_task_without_blocking_gui(
+            task=lambda: client.get_xpub('m', 'standard', creating=is_creating_wallet))
         client.used()
+        return client
 
     def get_xpub(self, device_id, derivation, xtype, wizard):
         if xtype not in self.SUPPORTED_XTYPES:
             raise ScriptTypeNotSupported(_('This type of script is not supported with {}.').format(self.device))
-        devmgr = self.device_manager()
-        client = devmgr.client_by_id(device_id)
-        client.handler = wizard
+        client = self.scan_and_create_client_for_device(device_id=device_id, wizard=wizard)
         xpub = client.get_xpub(derivation, xtype)
         client.used()
         return xpub
@@ -327,11 +320,11 @@ class TrezorPlugin(HW_PluginBase):
             return OutputScriptType.PAYTOMULTISIG
         raise ValueError('unexpected txin type: {}'.format(electrum_txin_type))
 
-    def sign_transaction(self, keystore, tx, prev_tx, xpub_path):
-        prev_tx = { bfh(txhash): self.electrum_tx_to_txtype(tx, xpub_path) for txhash, tx in prev_tx.items() }
+    def sign_transaction(self, keystore, tx: PartialTransaction, prev_tx):
+        prev_tx = { bfh(txhash): self.electrum_tx_to_txtype(tx) for txhash, tx in prev_tx.items() }
         client = self.get_client(keystore)
-        inputs = self.tx_inputs(tx, xpub_path, True)
-        outputs = self.tx_outputs(keystore.get_derivation(), tx)
+        inputs = self.tx_inputs(tx, for_sig=True, keystore=keystore)
+        outputs = self.tx_outputs(tx, keystore=keystore)
         details = SignTx(lock_time=tx.locktime, version=tx.version)
         signatures, _ = client.sign_tx(self.get_coin_name(), inputs, outputs, details=details, prev_txes=prev_tx)
         signatures = [(bh2u(x) + '01') for x in signatures]
@@ -343,7 +336,7 @@ class TrezorPlugin(HW_PluginBase):
         if not self.show_address_helper(wallet, address, keystore):
             return
         deriv_suffix = wallet.get_address_index(address)
-        derivation = keystore.derivation
+        derivation = keystore.get_derivation_prefix()
         address_path = "%s/%d/%d"%(derivation, *deriv_suffix)
         script_type = self.get_trezor_input_script_type(wallet.txin_type)
 
@@ -355,112 +348,107 @@ class TrezorPlugin(HW_PluginBase):
             sorted_pairs = sorted(zip(pubkeys, xpubs))
             multisig = self._make_multisig(
                 wallet.m,
-                [(xpub, deriv_suffix) for _, xpub in sorted_pairs])
+                [(xpub, deriv_suffix) for pubkey, xpub in sorted_pairs])
         else:
             multisig = None
 
         client = self.get_client(keystore)
         client.show_address(address_path, script_type, multisig)
 
-    def tx_inputs(self, tx, xpub_path, for_sig=False):
+    def tx_inputs(self, tx: Transaction, *, for_sig=False, keystore: 'TrezorKeyStore' = None):
         inputs = []
         for txin in tx.inputs():
             txinputtype = TxInputType()
-            if txin['type'] == 'coinbase':
+            if txin.is_coinbase_input():
                 prev_hash = b"\x00"*32
                 prev_index = 0xffffffff  # signed int -1
             else:
                 if for_sig:
-                    x_pubkeys = txin['x_pubkeys']
-                    xpubs = [parse_xpubkey(x) for x in x_pubkeys]
-                    multisig = self._make_multisig(txin.get('num_sig'), xpubs, txin.get('signatures'))
-                    script_type = self.get_trezor_input_script_type(txin['type'])
+                    assert isinstance(tx, PartialTransaction)
+                    assert isinstance(txin, PartialTxInput)
+                    assert keystore
+                    if len(txin.pubkeys) > 1:
+                        xpubs_and_deriv_suffixes = get_xpubs_and_der_suffixes_from_txinout(tx, txin)
+                        multisig = self._make_multisig(txin.num_sig, xpubs_and_deriv_suffixes)
+                    else:
+                        multisig = None
+                    script_type = self.get_trezor_input_script_type(txin.script_type)
                     txinputtype = TxInputType(
                         script_type=script_type,
                         multisig=multisig)
-                    # find which key is mine
-                    for xpub, deriv in xpubs:
-                        if xpub in xpub_path:
-                            xpub_n = parse_path(xpub_path[xpub])
-                            txinputtype.address_n = xpub_n + deriv
-                            break
+                    my_pubkey, full_path = keystore.find_my_pubkey_in_txinout(txin)
+                    if full_path:
+                        txinputtype.address_n = full_path
 
-                prev_hash = bfh(txin['prevout_hash'])
-                prev_index = txin['prevout_n']
+                prev_hash = txin.prevout.txid
+                prev_index = txin.prevout.out_idx
 
-            if 'value' in txin:
-                txinputtype.amount = txin['value']
+            if txin.value_sats() is not None:
+                txinputtype.amount = txin.value_sats()
             txinputtype.prev_hash = prev_hash
             txinputtype.prev_index = prev_index
 
-            if txin.get('scriptSig') is not None:
-                script_sig = bfh(txin['scriptSig'])
-                txinputtype.script_sig = script_sig
+            if txin.script_sig is not None:
+                txinputtype.script_sig = txin.script_sig
 
-            txinputtype.sequence = txin.get('sequence', 0xffffffff - 1)
+            txinputtype.sequence = txin.nsequence
 
             inputs.append(txinputtype)
 
         return inputs
 
-    def _make_multisig(self, m, xpubs, signatures=None):
+    def _make_multisig(self, m, xpubs):
         if len(xpubs) == 1:
             return None
-
         pubkeys = [self._make_node_path(xpub, deriv) for xpub, deriv in xpubs]
-        if signatures is None:
-            signatures = [b''] * len(pubkeys)
-        elif len(signatures) != len(pubkeys):
-            raise RuntimeError('Mismatched number of signatures')
-        else:
-            signatures = [bfh(x)[:-1] if x else b'' for x in signatures]
-
         return MultisigRedeemScriptType(
             pubkeys=pubkeys,
-            signatures=signatures,
+            signatures=[b''] * len(pubkeys),
             m=m)
 
-    def tx_outputs(self, derivation, tx):
+    def tx_outputs(self, tx: PartialTransaction, *, keystore: 'TrezorKeyStore'):
 
         def create_output_by_derivation():
-            script_type = self.get_trezor_output_script_type(info.script_type)
-            deriv = parse_path("/%d/%d" % index)
-            multisig = self._make_multisig(m, [(xpub, deriv) for xpub in xpubs])
+            script_type = self.get_trezor_output_script_type(txout.script_type)
+            if len(txout.pubkeys) > 1:
+                xpubs_and_deriv_suffixes = get_xpubs_and_der_suffixes_from_txinout(tx, txout)
+                multisig = self._make_multisig(txout.num_sig, xpubs_and_deriv_suffixes)
+            else:
+                multisig = None
+            my_pubkey, full_path = keystore.find_my_pubkey_in_txinout(txout)
+            assert full_path
             txoutputtype = TxOutputType(
                 multisig=multisig,
-                amount=amount,
-                address_n=parse_path(derivation + "/%d/%d" % index),
+                amount=txout.value,
+                address_n=full_path,
                 script_type=script_type)
             return txoutputtype
 
         def create_output_by_address():
             txoutputtype = TxOutputType()
-            txoutputtype.amount = amount
-            if _type == TYPE_SCRIPT:
-                txoutputtype.script_type = OutputScriptType.PAYTOOPRETURN
-                txoutputtype.op_return_data = trezor_validate_op_return_output_and_get_data(o)
-            elif _type == TYPE_ADDRESS:
+            txoutputtype.amount = txout.value
+            if address:
                 txoutputtype.script_type = OutputScriptType.PAYTOADDRESS
                 txoutputtype.address = address
+            else:
+                txoutputtype.script_type = OutputScriptType.PAYTOOPRETURN
+                txoutputtype.op_return_data = trezor_validate_op_return_output_and_get_data(txout)
             return txoutputtype
 
         outputs = []
         has_change = False
         any_output_on_change_branch = is_any_tx_output_on_change_branch(tx)
 
-        for o in tx.outputs():
-            _type, address, amount = o.type, o.address, o.value
+        for txout in tx.outputs():
+            address = txout.address
             use_create_by_derivation = False
 
-            info = tx.output_info.get(address)
-            if info is not None and not has_change:
-                index, xpubs, m = info.address_index, info.sorted_xpubs, info.num_sig
-                on_change_branch = index[0] == 1
+            if txout.is_mine and not has_change:
                 # prioritise hiding outputs on the 'change' branch from user
                 # because no more than one change address allowed
                 # note: ^ restriction can be removed once we require fw
                 # that has https://github.com/trezor/trezor-mcu/pull/306
-                if on_change_branch == any_output_on_change_branch:
+                if txout.is_change == any_output_on_change_branch:
                     use_create_by_derivation = True
                     has_change = True
 
@@ -472,17 +460,17 @@ class TrezorPlugin(HW_PluginBase):
 
         return outputs
 
-    def electrum_tx_to_txtype(self, tx, xpub_path):
+    def electrum_tx_to_txtype(self, tx: Optional[Transaction]):
         t = TransactionType()
         if tx is None:
             # probably for segwit input and we don't need this prev txn
             return t
-        d = deserialize(tx.raw)
-        t.version = d['version']
-        t.lock_time = d['lockTime']
-        t.inputs = self.tx_inputs(tx, xpub_path)
+        tx.deserialize()
+        t.version = tx.version
+        t.lock_time = tx.locktime
+        t.inputs = self.tx_inputs(tx)
         t.bin_outputs = [
-            TxOutputBinType(amount=vout['value'], script_pubkey=bfh(vout['scriptPubKey']))
-            for vout in d['outputs']
+            TxOutputBinType(amount=o.value, script_pubkey=o.scriptpubkey)
+            for o in tx.outputs()
         ]
         return t
